# xo-qemu laptop.
#
# Copyright (C) 2007 Mitchell N Charity
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os
import re
import shutil
import subprocess
import sys
import time
import ConfigParser

from laptop_init import ProtoLaptop
import qemu
from util import system_or_exception, getoutput_or_exception
from util import Ports
from util import XDisplay

class LaptopConfig:
    def create(self,fn):
        if os.path.exists(fn):
            raise Exception("File already exists: %s"%(fn))
        file(fn,"w").write("""
[Options]

memory: 256
# qemu -m 256

serial: vc
# qemu -serial vc
# vc|stdio|tty|none|...

soundhw: es1370
# qemu -soundhw es1370
# es1370|none

net_arguments: -net user -net nic,model=rtl8139
# qemu -net user -net nic,model=rtl8139

use_xephyr: False
# Start xephyr automatically.  False|True

config_file_format_version: 0 # Don't change.
""")
    create = classmethod(create)


    def __init__(self,fn):
        self._config = ConfigParser.ConfigParser()
        ret = self._config.read(fn)
        if len(ret) < 1:
            raise Exception("Failed to read config file %s"%(fn))

    def config(self):
        return self._config

    def get_option(self,option_key):
        return self.config().get('Options',option_key)


class Laptop:

    DISK_IMAGE_FILE = "disk.img"
    RUNTIME_DIR = "current/" 
    QEMU_PID_FILE = "current/qemu_pid"
    SSH_PORT_FILE = "current/ssh_port"
    NATIVE_X_XAUTH_FILE = "current/own_xauth"
    NATIVE_X_DISPLAY_FILE = "current/own_DISPLAY"
    XEPHYR_PID_FILE = "current/xephyr_pid"
    XEPHYR_XAUTH_FILE = "current/xephyr_xauth"
    XEPHYR_DISPLAY_FILE = "current/xephyr_DISPLAY"
    MOUNT_POINTS_FILE = "current/mount_points"
    CONFIG_DIR = "config/"
    SSH_KEY_FILE = "config/ssh_key"
    XEPHYR_COOKIE_FILE = "config/xephyr_cookie"
    CONFIG_FILE = "config/options.ini"
    XO_FILE = "xo"
    SETUP_REAL_XO_SCRIPT_FILE = "setup_real_xo_for_remote_access"

    REAL_XO_IP_ADDRESS_FILE = "ip_address"

    def make_laptop_from_build_file(self,laptop,build_file):
        proto = ProtoLaptop(laptop)
        proto.use_build_img_bz2(build_file)
        proto.initialize_protolaptop()
        return Laptop.new_from(laptop)

    make_laptop_from_build_file = classmethod(make_laptop_from_build_file)

    def make_laptop_remote(self,laptop):
        proto = ProtoLaptop(laptop)
        proto.initialize_protolaptop_for_remote_access()
        l = RealLaptop(laptop)
        l.set_ip('the_ip_address_has_not_been_set')
        return l

    make_laptop_remote = classmethod(make_laptop_remote)

    def new_from(self,laptop):
        ip_address = os.path.join(laptop,Laptop.REAL_XO_IP_ADDRESS_FILE)
        if os.path.exists(ip_address):
            return RealLaptop(laptop)
        else:
            return QemuLaptop(laptop)

    new_from = classmethod(new_from)

class QemuLaptop:

    def __init__(self, laptop):
        self.laptop = re.sub('\/*$',"",laptop)
        
    def _path(self, file):
        return os.path.join(self.laptop,file)

    def pid_file_is_active(self,pid_file):
        if not os.path.exists(pid_file):
            return False
        else:
            pid = file(pid_file).read()
            if os.system("/bin/ps %s > /dev/null"%(pid)) == 0:
                return True
            else:
                return False

    def is_running(self):
        pid_file = self._path(Laptop.QEMU_PID_FILE)
        running = self.pid_file_is_active(pid_file)
        return running

    def _cleanup_filesystem(self):
        self.mount_gc()
        runtime = self._path(Laptop.RUNTIME_DIR)
        if os.path.exists(runtime):
            shutil.rmtree(runtime)
            
    def power_on(self):
        if self.is_running():
            raise Exception("Laptop is already running: %s"%(self.laptop))
        self._cleanup_filesystem()

        ssh_port = str(Ports().find_unused_ssh_port())
        display = XDisplay(":%d"%(Ports().find_unused_X_display()))
        x_port = str(display.port())
        disk_img = self._path(Laptop.DISK_IMAGE_FILE)
        config = LaptopConfig(self._path(Laptop.CONFIG_FILE))

        soundhw = config.get_option('soundhw')
        if re.match('\s*none',soundhw) or re.match('\s+$',soundhw):
            sound_arguments = []
        else:
            sound_arguments = ['-soundhw',soundhw]
            
       	# If we have a qemu with kqemu support, we start it with kqemu support. Else without
        try: 
                cmd = ([qemu.Qemu().qemu_program(),
                        "-kernel-kqemu",# didn't work for me, because I didn't have it compiled with kqemu support. 
                        "-redir","tcp:"+ssh_port+"::22",
                        "-redir","tcp:"+x_port+"::6000",
                        "-m",config.get_option('memory'),
                        "-serial",config.get_option('serial')]
                       + sound_arguments
                       + config.get_option('net_arguments').split()
                       + ["-hda",disk_img])
                print >>sys.stderr, cmd

                self.power_on_setup_pre(ssh_port,display)

                pid = os.spawnvp(os.P_NOWAIT,cmd[0],cmd[1:])
        except:
                # qemu didn't start, so we try it withou kqemu.
                cmd = ([qemu.Qemu().qemu_program(),
                        "-redir","tcp:"+ssh_port+"::22",
                        "-redir","tcp:"+x_port+"::6000",
                        "-m",config.get_option('memory'),
                        "-serial",config.get_option('serial')]
                       + sound_arguments
                       + config.get_option('net_arguments').split()
                       + ["-hda",disk_img])
                print >>sys.stderr, cmd

                self.power_on_setup_pre(ssh_port,display)

                pid = os.spawnvp(os.P_NOWAIT,cmd[0],cmd[1:])
            
        file(self._path(Laptop.QEMU_PID_FILE),"w").write(str(pid))
        
        self.power_on_setup_post()

    def power_on_setup_pre(self,ssh_port,display):
        runtime = self._path(Laptop.RUNTIME_DIR)
        os.mkdir(runtime)
        file(self._path(Laptop.SSH_PORT_FILE),"w").write(ssh_port)
        file(self._path(Laptop.NATIVE_X_DISPLAY_FILE),"w").write("%s"%(display))

    def power_on_setup_post(self):
        config = LaptopConfig(self._path(Laptop.CONFIG_FILE))
        if config.get_option('use_xephyr')=='True':
            delay = 20
            print "Waiting %d seconds to start Xephyr..."%(delay)
            time.sleep(delay)
            self.xephyr_on()


    def power_off(self):
        self.xephyr_cleanup()
        try:
            pid_file = self._path(Laptop.QEMU_PID_FILE)
            pid = file(pid_file).read()
            os.system("kill %s"%(pid))
            os.unlink(pid_file)
        finally:
            self._cleanup_filesystem()

    def ssh_host_and_port(self):
        host = "localhost"
        port = file(self._path(Laptop.SSH_PORT_FILE)).read()
        return (host,port)

    def ssh(self,who,args,**kwargs):
        host,port = self.ssh_host_and_port()
        key_file = self._path(Laptop.SSH_KEY_FILE)
        cmd = ['ssh',
               '-o','NoHostAuthenticationForLocalhost=yes',
               '-p',port,
               '-i',key_file,
               who+"@"+host
               ] + args
        #print >>sys.stderr, cmd
        if 'getoutput' in kwargs:
            output = subprocess.Popen(cmd, stdout=subprocess.PIPE
                                      ).communicate()[0]
            return output
        else:
            #XXX check_call isn't working?!?
            status = subprocess.Popen(cmd).wait()
            if status != 0:
                raise Exception("ssh error(%d)"%(status))

    def scp(self,args):
        host,port = self.ssh_host_and_port()
        key_file = self._path(Laptop.SSH_KEY_FILE)
        args = [re.sub('^(olpc|root):',('\\1@'+host+':'),arg) for arg in args]
        cmd = ['scp',
               '-o', 'NoHostAuthenticationForLocalhost=yes',
               '-P',port,
               '-i',key_file
               ] + args
        #print >>sys.stderr, cmd
        status = subprocess.Popen(cmd).wait()
        if status != 0:
            raise Exception("ssh error(%d)"%(status))

    def get_xo_X_cookie(self):
        #TODO 2008-Jan-27 .Xauthority now looks valid.
        # Switch over the next time we discard back-compatibility.
        get_auth = 'F=`ls -t .serverauth.* | head -1`;echo list |xauth -f $F'
        auth = self.ssh('olpc',[get_auth],getoutput=True)
        pat = ':1\s+MIT-MAGIC-COOKIE-1\s+(\S+)'
        m = re.search(pat,auth)
        cookie = m.group(1)
        return cookie
        
    def update_runtime(self):
        cookie = self.get_xo_X_cookie()
        display = self.display()
        fn = self._path(Laptop.NATIVE_X_XAUTH_FILE)
        self.set_xauth_file_to(fn,display,cookie)

    def set_xauth_file_to(self,fn,display,cookie):
        if os.path.exists(fn):
            os.unlink(fn)
        cookie = cookie.rstrip("\n")
        cmd = "echo add %s . %s | xauth -f %s"%(display,cookie,fn)
        system_or_exception(cmd)
        

    def display(self):
        if self.xephyr_is_running():
            return file(self._path(Laptop.XEPHYR_DISPLAY_FILE)).read()
        else:
            return file(self._path(Laptop.NATIVE_X_DISPLAY_FILE)).read()

    def display_kludge(self):
        #XXX This may be a workaround for the machine's hostname
        #  not resolving properly.
        # Ie, display ":nnn" results in xauth using hostname/unix:nnn
        # but DISPLAY=:nnn xterm is then refused connection with that xauth.
        dpy = self.display()
        if re.match(':',dpy):
            return 'localhost'+dpy
        else:
            return dpy

    def xauth_file(self):
        if self.xephyr_is_running():
            return self._path(Laptop.XEPHYR_XAUTH_FILE)
        else:
            self.update_runtime()
            return self._path(Laptop.NATIVE_X_XAUTH_FILE)

    def xephyr_is_running(self):
        pid_file = self._path(Laptop.XEPHYR_PID_FILE)
        return self.pid_file_is_active(pid_file)

    def kill_any_xephyr(self):
        pid_file = self._path(Laptop.XEPHYR_PID_FILE)
        if os.path.exists(pid_file):
            os.system("kill `/bin/cat %s`"%(pid_file))
            time.sleep(0.9)
            os.system("kill -9 `/bin/cat %s`"%(pid_file))
            os.unlink(pid_file)

    def xephyr_on(self,display=None):
        if self.xephyr_is_running():
            raise Exception("Xephyr is already running on %s"%(self.laptop))
        if display:
            display = XDisplay(display)
        else:
            display = XDisplay(":%d"%(Ports().find_unused_X_display()))
        print "using %s"%(display)

        lang = self.ssh('olpc',["grep LANG= /home/olpc/.i18n"],getoutput=True)
        lang = lang.rstrip("\n")
        lang = re.sub('#.*',"",lang)

        cookie_file = self._path(Laptop.XEPHYR_COOKIE_FILE)
        cookie = file(cookie_file).read().strip()
        fn = self._path(Laptop.XEPHYR_XAUTH_FILE)
        self.set_xauth_file_to(fn,display,cookie)
        xauth_file = self._path(Laptop.XEPHYR_XAUTH_FILE)

        cmd = ["/bin/sh","-c",'%s exec Xephyr %s -auth %s -reset -screen 1200x900x16 -dpi 200'%(lang, display, xauth_file)]
        #print cmd

        xephyr_pid = subprocess.Popen(cmd).pid
        file(self._path(Laptop.XEPHYR_DISPLAY_FILE),'w').write("%s"%(display))
        file(self._path(Laptop.XEPHYR_PID_FILE),'w').write(str(xephyr_pid))
        
        if self.is_qemu_laptop():
            host = "10.0.2.2"
        else:
            host = getoutput_or_exception("/sbin/ifconfig eth0|perl -ne 'print $1 if /inet addr:([^ ]+)/'")
        display_on_xo = "%s:%d"%(host,display.number())
        self.ssh('olpc',["echo 'add %s . %s' | xauth"%(display_on_xo,cookie)])
        self.ssh('root',["telinit 3"])
        time.sleep(3)
        self.ssh('olpc',["DISPLAY=%s /home/olpc/remotedisplay &"%(display_on_xo)])

    def remove_file_if_it_exists(self,fn):
        if os.path.exists(fn):
            os.unlink(fn)

    def xephyr_cleanup(self):
        self.kill_any_xephyr()
        self.remove_file_if_it_exists(self._path(Laptop.XEPHYR_PID_FILE))
        self.remove_file_if_it_exists(self._path(Laptop.XEPHYR_XAUTH_FILE))
        self.remove_file_if_it_exists(self._path(Laptop.XEPHYR_DISPLAY_FILE))

    def xephyr_off(self):
        self.xephyr_cleanup()
        self.ssh('root',["telinit 5"])

    def sugar_restart(self):
        if self.xephyr_is_running():
            display = self.display()
            self.xephyr_cleanup()
            self.xephyr_on(display)
        else:
            self.ssh('root',['/usr/bin/killall /usr/bin/X'])
            self.update_runtime()

    def mount(self,what,where,user):
        # We need to give ssh arguments which sshfd isn't willing to pass.:(
        wrapper = getoutput_or_exception('/usr/bin/which sshfd_ssh_wrapper'
                                         ).rstrip()
        env = os.environ
        key_file = os.path.abspath(self._path(Laptop.SSH_KEY_FILE))
        host,port = self.ssh_host_and_port()
        more_ssh_args = (
            "-o NoHostAuthenticationForLocalhost=yes -i %s -p %s"
            %(key_file, port))
        env['SSHFD_SSH_WRAPPER_ARGUMENTS'] = more_ssh_args

        where_abs = os.path.abspath(where)
        file(self._path(Laptop.MOUNT_POINTS_FILE),'a').write(where_abs+"\n")

        cmd = ['sshfs',
               '-o','intr,direct_io,cache=no,sync_read,reconnect,TCPKeepAlive=yes',
               '-o',"ssh_command=%s"%(wrapper),
               '%s@%s:%s'%(user,host,what),
               where]
        system_or_exception(" ".join(cmd))

    def umount(self,where):
        system_or_exception("fusermount -u %s"%(where))

    def mount_gc(self): #XXX rename, reconsider
        if not os.path.exists(self._path(Laptop.RUNTIME_DIR)):
            return
        fn = self._path(Laptop.MOUNT_POINTS_FILE)
        if os.path.exists(fn):
            mounts = [line.rstrip() for line in file(fn).readlines()]
            for mount in mounts:
                if os.path.ismount(mount):
                    self.umount(mount)
        file(fn,'w').write("")

    def is_qemu_laptop(self):
        return True


class RealLaptop (QemuLaptop):

    def is_running(self):
        host,port = self.ssh_host_and_port()
        ping = os.system("/bin/ping -c 1 -W 1 -q %s > /dev/null"%(host)) == 0
        if ping:
            return True
        else:
            return False

    def power_on(self):
        host,port = self.ssh_host_and_port()
        self.power_on_setup_pre('22',host+':0')
        self.power_on_setup_post()

    def power_off(self):
        self._cleanup_filesystem()
        self.xephyr_cleanup()

    def ssh_host_and_port(self):
        host = file(self._path(Laptop.REAL_XO_IP_ADDRESS_FILE)).read().strip()
        port = '22'
        return (host,port)

    def set_ip(self,ip_address):
        ip_address = ip_address.strip()
        file(self._path(Laptop.REAL_XO_IP_ADDRESS_FILE),'w').write(ip_address)

    def is_qemu_laptop(self):
        return False
