# Create and setup an xo-qemu laptop.
#
# Copyright (C) 2007 Mitchell N Charity
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import re
import sys
import time

import olpc_xo_qemu.pexpect as pexpect

from qemu import Qemu
import laptop
from util import system_or_exception, getoutput_or_exception

class ProtoLaptop:

    def __init__(self,directory):
        if os.path.exists(directory):
            raise Exception("Directory already exists: %s"%(directory))
        self.dir = re.sub('\/*$',"/",directory)
        os.mkdir(self.dir)
        self.disk_image = self.dir+laptop.Laptop.DISK_IMAGE_FILE
        self.xo = None

    def use_build_img_bz2(self,img_bz_file):
        # qemu-img CVS qcow2 files are not back compatible with released qemu.
        #DiskImageManipulation().convert_img_bz2_to_qcow2(img_bz_file, self.disk_image)
        system_or_exception("bzcat %s > %s"%(img_bz_file,self.disk_image))
        
    def initialize_protolaptop_for_remote_access(self):
        config = self.dir+laptop.Laptop.CONFIG_DIR
        os.mkdir(config) # _not_ in the try.
        try:
            self.setup_ssh_and_xauth_files()
        except:
            if os.path.exists(config):
                shutil.rmtree(config)
        self.create_config_file()
        self.create_xo_file()
        self.create_real_xo_setup_script()

    def initialize_protolaptop(self):
        self.startup_qemu()
        config = self.dir+laptop.Laptop.CONFIG_DIR
        runtime = self.dir+laptop.Laptop.RUNTIME_DIR
        os.mkdir(config) # _not_ in the try.
        try:
            os.mkdir(runtime)
            self.setup_ssh_and_xauth_files()
            self.setup_ssh()
            self.setup_xauth_for_xephyr()
            self.setup_remotedisplay_script()
            #self.patch_bashrc_with_sugar_environment_variables()
            #self.patch_home_olpc_permissions_for_sshd()
            self.patch_enable_XTEST()
            self.shutdown_qemu()
        except:
            try:
                self.shutdown_qemu()
            except:
                pass
            if os.path.exists(config):
                shutil.rmtree(config)
            if os.path.exists(runtime):
                shutil.rmtree(runtime)
            os.unlink(self.dir+laptop.Laptop.DISK_IMAGE_FILE)
            raise
        self.create_config_file()
        self.create_xo_file()

    def create_config_file(self):
        laptop.LaptopConfig.create(self.dir+laptop.Laptop.CONFIG_FILE)

    def create_xo_file(self):
        fn = self.dir+laptop.Laptop.XO_FILE
        contents = """#!/bin/sh\nexec xo-qemu --xo-api0 "$0" "$@"\n"""
        file(fn,"w").write(contents)
        os.chmod(fn,0755)

    def startup_qemu(self):
    	# If we have a qemu with kqemu support, we start it with kqemu support. Else without. 
	try: 
                cmd = (Qemu().qemu_program()+
                        " -kernel-kqemu" # removed, because it wasn't recognized by my qemu (threw an error). 
                        " -serial stdio"
                        " -m 256 -net user -net nic,model=rtl8139"
                        " -hda "+self.disk_image)
                print >>sys.stderr, cmd
                self.xo = pexpect.spawn(cmd)
	except: 
	        # if that didn't work, call qemu withou kqemu. 
		cmd = (Qemu().qemu_program()+
			" -serial stdio"
			" -m 256 -net user -net nic,model=rtl8139"
			" -hda "+self.disk_image)
	        print >>sys.stderr, cmd
		self.xo = pexpect.spawn(cmd)

        #self.xo.expect("login:",timeout=60*5)
        # 2008-Jan-27 - On joyride, the login prompt no longer appears
        # without pressing return. :(
        elapsed = 0
        timeout_after = 60*5
        hit_return_every = 20
        while True:
            try:
                which = self.xo.expect(["login:","Password:"],
                                       timeout=hit_return_every)
                if which == 1: # Too many returns.
                    self.xo.send("\n")
                    continue
                break
            except pexpect.TIMEOUT:
                elapsed += hit_return_every
                if elapsed >= timeout_after:
                    raise
                self.xo.send("\n")

        self.xo.send("root\n")
        index = self.xo.expect(["Password","bash","olpc@"])
        if index == 0:
            self.xo.send("\n")
            self.xo.expect("login:")
            self.xo.send("olpc\n")
            self.expect_prompt()
            self.xo.send("sudo -i\n")
            self.expect_prompt()
            
    def shutdown_qemu(self):
        self.xo.send("\n echo -n foo1; echo bar2\n")
        self.xo.expect("foo1bar2")
        self.xo.send("/sbin/shutdown -hP now\n")
        time.sleep(12)
        self.xo.close()
        self.xo.kill(1)
        self.xo = None

    def need_root_shell(self):
        self.xo.send("\n")
        self.xo.expect(["bash"])
        self.xo.send("cd\n")

    def expect_prompt(self):
        self.xo.expect(["bash","olpc@"])
        
    def setup_ssh_and_xauth_files(self):
        ssh_key_file = self.dir+laptop.Laptop.SSH_KEY_FILE
        system_or_exception("ssh-keygen -N '' -t rsa -f "+ssh_key_file)

        cookie_file = self.dir+laptop.Laptop.XEPHYR_COOKIE_FILE
        system_or_exception("mcookie > %s"%(cookie_file))

    def setup_ssh(self):
        ssh_key_file = self.dir+laptop.Laptop.SSH_KEY_FILE
        public_key = file(ssh_key_file+".pub").read()

        self.need_root_shell()
        self.xo.send("gawk '/EOF EOF/ {exit} {print}' > /tmp/id_rsa.pub\n")
        time.sleep(2)
        self.xo.send(public_key+"\n")
        self.xo.send("EOF EOF\n")
        self.expect_prompt()
        self.xo.send("""
        mkdir ~root/.ssh
        cat /tmp/id_rsa.pub > ~root/.ssh/authorized_keys
        chmod go-rwx ~root/.ssh ~root/.ssh/authorized_keys
        su olpc
        mkdir ~olpc/.ssh
        cat /tmp/id_rsa.pub > ~olpc/.ssh/authorized_keys
        chmod go-rwx ~olpc/.ssh ~olpc/.ssh/authorized_keys
        exit
        rm /tmp/id_rsa.pub
        """)

    def setup_xauth_for_xephyr(self):
        cookie_file = self.dir+laptop.Laptop.XEPHYR_COOKIE_FILE
        cookie = file(cookie_file).read()

        self.need_root_shell()
        self.xo.send("""
        su olpc
        cd
        """)

    def setup_remotedisplay_script(self):
        # http://wiki.laptop.org/go/Remote_display#Remote_Sugar_session
        self.need_root_shell()
        self.xo.send("""
        su olpc
        cd
        cp /usr/bin/olpc-session remotedisplay
        chmod a+x remotedisplay
        exit
        """)

    def patch_bashrc_with_sugar_environment_variables(self):
        self.need_root_shell()
        self.xo.send("""
        su olpc
        cd
        grep export /usr/bin/sugar >> .bashrc
        exit
        """)

    def patch_home_olpc_permissions_for_sshd(self):
        # http://dev.laptop.org/ticket/5320
        # Believed unneeded joyride >= 1508
        self.need_root_shell()
        self.xo.send("/bin/chmod g-w /home/olpc\n")
        
    def patch_enable_XTEST(self):
        self.need_root_shell()
        self.xo.send("""
        cd /etc/X11
        cp -p xorg-emu.conf xorg-emu.conf.bak
        gawk '/"XTEST"/ {print "  Option \\"XTEST\\" \\"Enable\\" \# set by xo-qemu"} !/"XTEST"/ {print}' xorg-emu.conf.bak > xorg-emu.conf
        """)


    def create_real_xo_setup_script(self):
        ssh_key_file = self.dir+laptop.Laptop.SSH_KEY_FILE
        public_key_file = ssh_key_file + ".pub"
        script = ("""#!/bin/env python
import os
import sys
import commands
import re

class Setup:
    def __init__(self):
        laptop = sys.argv[0]
        if not os.path.isdir(laptop):
            laptop = os.path.dirname(laptop)
        self.laptop = laptop

    def setup(self):
        self.assure_running_as_root()
        self.setup_ssh()
        self.patch_enable_XTEST()
        self.get_ip_address()
        self.system("/bin/sync")
        
    def assure_running_as_root(self):
        status,output = commands.getstatusoutput("/usr/bin/whoami")
        if output != "root":
            raise "Must be run as root."

    def system(self,command):
        status = os.system(command)
        if status != 0:
           raise Exception("Command failed(%d): %s"%(status,command))

    def get_ip_address(self):
        status,output = commands.getstatusoutput("/sbin/ifconfig eth0")
        m = re.search('inet addr:([^ ]+)',output)
        if not m:
           print "Failed to get ip address."
           print "You will have to configure it manually later."
        else:
           address = m.group(0)
           fn = os.path.join(self.laptop, """
                  + '"' + laptop.Laptop.REAL_XO_IP_ADDRESS_FILE + '"' + """)
           file(fn,"w").write(address)

    def setup_ssh(self):
        public_key_file = os.path.join(self.laptop,"""
                  + '"' + laptop.Laptop.SSH_KEY_FILE  + '.pub"' + """)
        cmd = ("\""
        cat "\"" + public_key_file + "\"" > /tmp/id_rsa.pub
        mkdir ~root/.ssh
        cat /tmp/id_rsa.pub > ~root/.ssh/authorized_keys
        chmod go-rwx ~root/.ssh ~root/.ssh/authorized_keys
        mkdir ~olpc/.ssh
        cat /tmp/id_rsa.pub > ~olpc/.ssh/authorized_keys
        chmod go-rwx ~olpc/.ssh ~olpc/.ssh/authorized_keys
        chown olpc:olpc ~olpc/.ssh ~olpc/.ssh/authorized_keys
        rm /tmp/id_rsa.pub
        "\"") 
        print cmd
        self.system(cmd)

    def patch_enable_XTEST(self):
        cmd = ("\""
        cd /etc/X11
        cp -p xorg-dcon.conf xorg-dcon.conf.bak
        gawk '/"XTEST"/ {print "  Option \\\\"XTEST\\\\" \\\\"Enable\\\\""} !/"XTEST"/ {print}' xorg-dcon.conf.bak > xorg-dcon.conf
        "\"")
        print cmd
        self.system(cmd)


Setup().setup()

""")
        fn = self.dir+laptop.Laptop.SETUP_REAL_XO_SCRIPT_FILE
        file(fn,"w").write(script)
        os.chmod(fn,0755)
