# xo-qemu script support.
# Script for downloading, managing, and running, OLPC QEMU disk images.
#
# Copyright (C) 2007 Mitchell N Charity
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import re
import sys
from sys import exit

import build_cache
import build_system
import disk_images
import laptop
from laptop import Laptop
import new_user

def xo_qemu_main(argv):
    #print >>sys.stderr, argv
    XoQemuScript().main(argv)

class XoQemuScript:

    def __init__(self):
        self.usage = """
xo-qemu is a script for downloading, managing, and running OLPC QEMU
disk images on linux.

The One Laptop Per Child (OLPC)<http://wiki.laptop.org/> project
creates a linux distribution for its laptop.  It is released as
numbered builds (ie, "versions"), in several streams (eg, separate
streams for developers and end users).  These builds are available in
several forms, one of which is a disk image that can be run on the
QEMU virtual machine.  This script makes it easier to obtain, manage,
alter, and run these disk images on linux.  It provides virtual,
slightly broken, emulated OLPC XO laptops.

"""
        self.usage_for = {}
        self.command = {}
        self.api0_command = {}
        def cmd(name,argl,method,desc):
            desc = re.sub('^\n','',desc)
            section = "--%s %s\n\n%s\n"%(name,argl,desc)
            self.usage += section
            self.usage_for[name] = section
            self.command[name] = method
        def api0(name,argl,method,desc):
            desc = re.sub('^\n','',desc)
            section = "  %s %s\n\n%s\n"%(name,argl,desc)
            self.usage += section
            self.usage_for[name] = section
            self.usage_for['xo-api0'] += section
            self.api0_command[name] = method


        cmd("help","", self.cmd_help, """
        Shows this help.
        """)

        cmd("new-user","", self.cmd_new_user, """
        Checks that qemu and kqemu are working,
        and suggests how to get started.
        """)

        cmd("report","", self.cmd_report, """
        Summarizes the currently available streams and builds.

        (The report format is likely to change).
        """)

        cmd("get-build","STREAM BUILD DIRECTORY", self.cmd_get_build, """
        Downloads STREAM's BUILD disk image into DIRECTORY.
        The build's devel_ext3/ .img.bz2 disk image is used.
        If there is already a copy of the image in DIRECTORY,
        it is not downloaded again.  
        The filename of the image in DIRECTORY is printed.

        BUILD can be "latest" instead of a number.

        Example:
        $ xo-qemu --get-build joyride latest builds/
        builds/olpc-redhat-stream-joyride-build-1418-20071213_1608-devel_ext3.img.bz2
        $ xo-qemu --get-build joyride 1418  builds/
        builds/olpc-redhat-stream-joyride-build-1418-20071213_1608-devel_ext3.img.bz2
        """)

        cmd("make-laptop","LAPTOP DISK_IMAGE", self.cmd_make_laptop, """
        Creates a new laptop named LAPTOP, from a build DISK_IMAGE.

        A new directory named LAPTOP is created, and must not already exist.
        The DISK_IMAGE is copied into it.  Each laptop is about 1 GB.

        As part of setting up the laptop, qemu is automatically run,
        some configuration done, and then qemu is automatically shut down.
        No user action is needed.

        The LAPTOP/xo file is used to control it.

        Examples:
        $ xo-qemu --make-laptop laptops/one builds/olpc-redhat-stream-joyride-build-1418-20071213_1608-devel_ext3.img.bz2
        $ xo-qemu --make-laptop laptops/one `xo-qemu --get-build joyride latest builds/ |tail -1`
        And then:
        $ laptops/one on
        """)

        cmd("xo-api0","LAPTOP ACTION", self.cmd_xo_api0, """
        Controls the LAPTOP.  This is usually called indirectly
        using LAPTOP/xo ACTION.
        
        Examples:
        $ laptops/one/xo on
        $ laptops/one/xo off
        """)

        api0("on","", self.api0_on, """
        """)
        api0("off","", self.api0_off, """
        """)
        api0("is-running","", self.api0_running, """
        """)

        api0("ssh","[olpc|root] [COMMAND-LINE]", self.api0_ssh, """
        """)
        api0("scp","[SCP_ARGS] FROM TO", self.api0_scp, """
        Copies FROM to TO, where both are either a local path,
        or [olpc|root]:path/on/the/laptop.
        """)

        api0("env","", self.api0_env, """
        Prints "LAPTOP=xxx DISPLAY=xxx XAUTHORITY=xxx".
        These environment variables let you run X windows programs
        on the laptop's X server, from the host.

        This is mostly used for automated testing.

        Examples:
        $ bash     # start up a new shell, since we're about to modifiy it
        $ export `laptops/one/xo env`
        $ xterm    # somewhat useless, but illustrates the concept.
        $ import screenshot_of_laptop.png
        """)
        api0("do","COMMAND [ARGS]", self.api0_do, """
        Sets up the same environment variables as /xo env,
        and then runs COMMAND.
        If and environment variable XO_QEMU_DO_PATH is defined,
        it is used instead of PATH to find COMMAND.

        Examples:
        $ laptops/one/xo do xterm
        $ laptops/one/xo do sh -c 'echo $DISPLAY'
        """)

        api0("xephyr","on|off", self.api0_xephyr, """
        Switches between using the laptop's normal X server,
        and using an external Xephyr X server.

        The real OLPC laptop hardware has a 1200x900 display,
        Until qemu supports having a 1200x900 display (the release after 0.9.0),
        Xephyr can be used when one is needed.

        Example:
        $ laptops/one/xo on
        $ laptops/one/xo xephyr on
        """)

        api0("sugar","restart", self.api0_sugar, """
        Restarts sugar.  This is needed after some configuration changes.
        """)

        api0("mount","[WHAT WHERE USER] | -u [WHERE]", self.api0_mount, """
        mount [WHAT WHERE USER]

        Mounts laptop directory WHAT on host directory WHERE,
        with the host user mapped to laptop USER olpc or root.

        If no arguments are given,
         / is mounted on LAPTOP/root with the user as root, and
         /home/olpc is mounted on LAPTOP/olpc with the user as olpc.

        mount -u WHERE

        Unmounts whatever is mounted on WHERE.
        Equivalent to "fusermount -u WHERE".

        If no WHERE is given, then LAPTOP/root and LAPTOP/olpc are unmounted.

        "/xo off" unmounts everything mounted.

        Example:
        $ laptops/one/xo on
        $ laptops/one/xo mount
        $ ls laptops/one/root
        activities  bin  boot  dev...
        $ ls -a laptops/one/olpc
        .AbiSuite .bash_history ... .sugar ...
        $ laptops/one/xo mount -u
        """)

        cmd("enlarge-disk-image","IMAGE SIZE_MB", self.cmd_enlarge_disk_image, """
        Enlarges the LAPTOP's disk to MB megabytes.
        The laptop's disk is normally 1000 MB, with only 1/3 used.
        But when doing development, and installing lots of things like gcc,
        it sometimes helps to make it bigger.
        """)

        cmd("make-laptop-remote","LAPTOP", self.cmd_make_laptop_remote, """
        Like make-laptop, but for remote access to a real XO laptop.

        Example steps:
 
        (1) Create a remote access laptop:
          $ xo-qemu --make-laptop-remote laptops/two
 
        (2) Move the laptop to a usb drive.
          $ mv laptops/two /media/whatever/two
          $ sync

        (3) On the real XO laptop, run setup_real_xo_for_remote_access, as root.
        [olpc@xo-...]$ sudo /media/something/two/setup_real_xo_for_remote_access
        [olpc@xo-...]$ sync

        (4) Back on the host, move the laptop back from the usb drive.
          $ mv /media/whatever/two laptops/two
          $ sync

        The laptop should now work:
          $ laptops/two/xo ssh whoami

        When the real XO laptop gets a new ip address from dhcp,
        you tell the remote access laptop about it by:
          $ laptops/two/xo ip 192.whatever
        """)

        api0("ip","ADDRESS", self.api0_ip, """
        Set the IP address on a remote access laptop.

        It is an error to request this action on a QEMU laptop.
        """)

        self.laptop = None # used by api0

    def print_usage_and_exit(self,status=2):
        print self.usage
        exit(status)

    def print_command_usage_and_exit(self,cmd,msg,status=2):
        print >>sys.stderr, msg,"\n\nUse --help to see all help.\n\n",
        print >>sys.stderr, self.usage_for[cmd]
        exit(status)

    def main(self,argv):
        if len(argv) < 2:
            self.print_usage_and_exit()
        cmd = argv[1]
        cmd = re.sub('^--','',cmd)
        if cmd not in self.command:
            print >>sys.stderr, "Unknown command: %s\n"%(cmd)
            self.print_usage_and_exit()
        self.command[cmd](argv[2:])
        exit(0)

    def assert_argv_len_in(self,cmd,lengths,argv):
        if not len(argv) in lengths:
            self.print_command_usage_and_exit(cmd,"Invalid argument count.")
    def assert_argv_len_minimum(self,cmd,minlen,argv):
        if not len(argv) >= minlen:
            self.print_command_usage_and_exit(cmd,"Invalid argument count.")

    def cmd_help(self,argv):
        self.assert_argv_len_in("report",(0,),argv)
        self.print_usage_and_exit(0)

    def cmd_new_user(self,argv):
        self.assert_argv_len_in("new-user",(0,),argv)
        new_user.NewUser().help_new_user_setup()
        exit(0)

    def cmd_report(self,argv):
        self.assert_argv_len_in("report",(0,),argv)
        streams = build_system.OLPCStreams.cached
        txt = ""
        for stream in streams.streams():
            latest = stream.latest_build()
            txt += "\n%s\n--------\n%s\nLatest is %s (%s).\n%s\n"%(
                stream.name, stream.url,
                latest.build_id, latest.modification_time().isoformat(),
                latest.urls_of_ext3_image_and_its_md5()[0])
        print txt

    def cmd_get_build(self,argv):
        self.assert_argv_len_in("get-build",(3,),argv)
        stream_name,build_id,dir = argv
        streams = build_system.OLPCStreams.cached
        stream = streams.get_stream_by_name(stream_name)
        build = stream.get_build_by_id(build_id)
	# If the dir doesn't exist, create it, because it is far more convenient for the user. 
	if not self.dir_exists(dir): 
	    print "created dir", self.create_dir(dir) # TODO: Correct text to no longer say thatc reating the dir is necessary.
        print build_cache.ObtainBuilds(dir).download_build_cached(build)

    def dir_exists(self, dir): 
        """Check if the dir exists."""
	from os.path import isdir
	return isdir(dir)

    def create_dir(self, dir): 
        """Create the dir. 
	
	This only works if a single dir was given."""
	from os import mkdir
	# First strip any possibly given path seperators from the dir
	dir = dir.replace("/", "")
	# now create the dir. 
	mkdir(dir)
	return dir



    def cmd_make_laptop(self,argv):
        self.assert_argv_len_in("make-laptop",(2,),argv)
        laptop,build_file = argv
        print (
            "\nAfter a minute or two, the laptop will start up,\n"
            "run briefly to configure itself, and then shutdown.\n"
            "You can then run it by saying: " + laptop + "/xo on\n\n")
        Laptop.make_laptop_from_build_file(laptop,build_file)

    def cmd_make_laptop_remote(self,argv):
        self.assert_argv_len_in("make-laptop-remote",(1,),argv)
        laptop, = argv
        Laptop.make_laptop_remote(laptop)
        print (
            "\nNow move " + laptop + " to the real XO,\n"
            "run " + laptop + "/setup_real_xo_for_remote_access\n"
            "and then move the laptop back.\n\n"
            "When the real XO's ip address later changes,\n"
            "run " + laptop + "/xo ip 192.whatever .\n\n")

    def cmd_xo_api0(self,argv):
        #print >>sys.stderr, argv
        self.assert_argv_len_minimum("xo-api0",2,argv)
        laptop = argv[0]
        if not os.path.isdir(laptop):
            laptop = os.path.dirname(laptop)
        self.laptop = laptop
        cmd = argv[1]
        if cmd not in self.api0_command:
            self.print_command_usage_and_exit(
                "xo-api0",
                "Unknown command: %s"%(cmd))
        self.api0_command[cmd](argv[2:])
        exit(0)

    def api0_on(self,argv):
        self.assert_argv_len_in("on",(0,),argv)
        laptop = Laptop.new_from(self.laptop)
        laptop.power_on()

    def api0_off(self,argv):
        self.assert_argv_len_in("off",(0,),argv)
        laptop = Laptop.new_from(self.laptop)
        if not laptop.is_running():
            print >>sys.stderr,(
                "ERROR: Laptop is not running.  laptop: %s"%(self.laptop))
            exit(1)
        laptop.power_off()

    def api0_running(self,argv):
        self.assert_argv_len_in("is-running",(0,),argv)
        if Laptop.new_from(self.laptop).is_running():
            print True
            exit(0)
        else:
            print False
            exit(1)

    def api0_ssh(self,argv):
        who = "olpc"
        args = argv
        if len(argv) >= 1 and argv[0] in ("root","olpc"):
            who = argv[0]
            args = argv[1:]
        Laptop.new_from(self.laptop).ssh(who,args)

    def api0_scp(self,argv):
        self.assert_argv_len_in("scp",(2,3,),argv)
        Laptop.new_from(self.laptop).scp(argv)

    def api0_env(self,argv):
        self.assert_argv_len_in("env",(0,),argv)
        str,ignore = self.determine_env()
        print str
        exit(0)

    def api0_do(self,argv):
        DO_PATH = 'XO_QEMU_DO_PATH'
        self.assert_argv_len_minimum("do",1,argv)
        ignore,info = self.determine_env()
        env = os.environ
        for k in info:
            env[k] = info[k]
        if DO_PATH in env and env[DO_PATH] != "":
            #env['PATH'] = "%s:%s"%(env[DO_PATH],env['PATH'])
            env['PATH'] = env[DO_PATH]
        try:
            os.execvpe(argv[0],argv,env)
        except OSError,e:
            print >>sys.stderr, argv,"\n",e

    def determine_env(self):
        class Env:
            def __init__(self):
                self.str = ""
                self.dict = {}
            def add(self,k,v):
                self.str += "%s=%s "%(k,v)
                self.dict[k] = v
        l = Laptop.new_from(self.laptop)
        e = Env()
        e.add('LAPTOP',self.laptop)
        e.add('DISPLAY',l.display_kludge())
        e.add('XAUTHORITY',l.xauth_file())
        return e.str,e.dict

    def api0_xephyr(self,argv):
        self.assert_argv_len_minimum("xephyr",1,argv)
        scmd = argv[0]
        if scmd == 'on':
            Laptop.new_from(self.laptop).xephyr_on()
        elif scmd == 'off':
            Laptop.new_from(self.laptop).xephyr_off()
        elif scmd == 'is-running':
            if Laptop.new_from(self.laptop).xephyr_is_running():
                print True
                exit(0)
            else:
                print False
                exit(1)
        else:
            self.print_command_usage_and_exit(
                "xo-api0",
                "Unknown sub-command: %s %s"%('xephyr',scmd))

    def api0_sugar(self,argv):
        self.assert_argv_len_minimum("sugar",1,argv)
        scmd = argv[0]
        if scmd == 'restart':
            Laptop.new_from(self.laptop).sugar_restart()
        else:
            self.print_command_usage_and_exit(
                "xo-api0",
                "Unknown sub-command: %s %s"%('sugar',scmd))

    def api0_mount(self,argv):
        #XXX too complex?
        self.assert_argv_len_in("mount",(0,1,2,3),argv)
        laptop = Laptop.new_from(self.laptop)
        default_root = self.laptop+"/"+"root"
        default_olpc = self.laptop+"/"+"olpc"
        def setup_dir(dir):
            if os.path.ismount(dir):
                laptop.umount(dir)
            if not os.path.isdir(dir):
                os.mkdir(dir)
        if len(argv)==0:
            setup_dir(default_root)
            setup_dir(default_olpc)
            laptop.mount('/',default_root,'root')
            laptop.mount('/home/olpc',default_olpc,'olpc')
        elif len(argv)==1 and argv[0] == '-u':
            if os.path.ismount(default_root):
                laptop.umount(default_root)
            if os.path.isdir(default_root):
                os.rmdir(default_root)
            if os.path.ismount(default_olpc):
                laptop.umount(default_olpc)
            if os.path.isdir(default_olpc):
                os.rmdir(default_olpc)
        elif len(argv)==1 and argv[0] == 'gc': #XXX not yet public
            laptop.mount_gc()
        elif len(argv)==2 and argv[0] == '-u':
            laptop.umount(argv[1])
        elif len(argv)==3:
            what,where,user = argv
            laptop.mount(what,where,user)
        else:
            self.print_command_usage_and_exit("mount","Invalid arguments.")
        exit(0)

    def api0_ip(self,argv):
        self.assert_argv_len_in("ip",(1,),argv)
        ip, = argv
        laptop = Laptop.new_from(self.laptop)
        laptop.set_ip(ip)

    def cmd_enlarge_disk_image(self,argv):
        self.assert_argv_len_in("enlarge-disk-image",(2,),argv)
        img,size = argv
        size = int(size)
        disk_images.DiskImageManipulation().enlarge_disk_image(img,size)
        exit(0)

