# Describe the OLPC build streams.
#
# Copyright (C) 2007 Mitchell N Charity
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import os
import re
import time

from util import wget_page

class OLPCStreams:

    db = None

    def __init__(self):
        self.base_urls = ["http://xs-dev.laptop.org/~cscott/olpc/streams/",
                          "http://pilgrim.laptop.org/~pilgrim/olpc/streams/"]
        self._scrape_a_stream = '>([\w\.]+)/</a>\s+(\d\S+\s+\d\d:\d\d)'

        self._initialized = False
        self._streams = None

        #FUTURE- it would be nice if the server had a README describing
        #the streams. Or one per stream.

    def _parse_time(self,str):
        return datetime.datetime(*time.strptime(str, "%d-%b-%Y %H:%M")[0:5])

    def _initialize(self):
        if not self._initialized:
            def get_site(url):
                page = wget_page(url)
                scrape_tuples = re.findall(self._scrape_a_stream, page)

                # cscott's copy of update.1 is now obsolete.
                if re.search('cscott',url):
                    scrape_tuples = [t for t in scrape_tuples
                                     if t[0] != "update.1"]

                def mkstream(scrape_tuple):
                    name,mod_t = scrape_tuple
                    return Stream(name, os.path.join(url,name),
                                  self._parse_time(mod_t))
                self._streams += [mkstream(tuple) for tuple in scrape_tuples]

            self._streams = []
            [get_site(url) for url in self.base_urls]
            self._initialized = True

    def streams(self):
        self._initialize()
        return self._streams

    def get_stream_by_name(self,name):
        for stream in self.streams():
            if stream.name == name:
                return stream
        return None

OLPCStreams.cached = OLPCStreams()

class Stream:
    
    def __init__(self,name,url,modification_time):
        self.name = name
        self.url = url
        self.modification_time = modification_time

        self._scrape_a_build = '>(\w+?(\d[\d\.a-z_]*)-(20[01][\d_]+))/\.?</a>'

        self._initialized = False
        self._builds = []
        self._build_idx = {}
        self._latest_build = None

    def _parse_time(self,str):
        return datetime.datetime(*time.strptime(str, "%Y%m%d_%H%M")[0:5])

    def _initialize(self):
        if not self._initialized:
            self._initialized = True
            page = wget_page(self.url)
            # builds
            builds = []
            scrape_tuples = re.findall(self._scrape_a_build, page)
            for scrape_tuple in scrape_tuples:
                dir_name, id, mod_t = scrape_tuple
                when = self._parse_time(mod_t)
                url = os.path.join(self.url,dir_name)
                b = Build(self.name,id,url,when)
                self._build_idx[id] = b
                builds.append(b)
            self._builds = builds
            # check
            if len(builds) == 0:
                raise Exception("%s page was not recognizable"%(self.url))
            # latest
            m = re.search('LATEST-IS-(\d[\d\._a-z]+)', page)
            if m:
                id = m.group(1)
                latest = self.get_build_by_id(id)
            else:
                latest = builds[-1]
            self._latest_build = latest

    def builds(self):
        self._initialize()
        return self._builds

    def get_build_by_id(self,id):
        self._initialize()
        if id == 'latest':
            return self.latest_build()
        else:
            return self._build_idx[id]

    def latest_build(self):
        self._initialize()
        return self._latest_build


class BuildDiskImageFile:

    def __init__(self):
        self._pat = '-([a-z0-9\._]+)-build-(\d[\d\._a-z]+)-.+\.img\.bz2$'

    def build_name_from_filename(self,filename):
        m = re.search(self._pat,filename)
        if m:
            stream_name = m.group(1)
            build_id = m.group(2)
            return BuildName(stream_name,build_id)
        else:
            return None
    build_name_from_filename = classmethod(build_name_from_filename)


class Build:

    def new_from_ext3_image_filename(self, filename):
        pat = '-([a-z0-9\._]+)-build-(\d[\d\._a-z]+)-.+\.img\.bz2$'
        m = re.search(pat,filename)
        stream_name = m.group(1)
        build_id = m.group(2)
        return Build(stream_name,build_id)
    new_from_ext3_image_filename = classmethod(new_from_ext3_image_filename)

    def __init__(self, stream_name, build_id,
                 url=None, modification_time=None):

        self.stream_name = stream_name
        self.build_id = build_id
        
        self._url = url
        self._modification_time = modification_time

        self._ext3_url_was_looked_for = False
        self._ext3_url = None

    def name(self):
        return "%s-%s"%(self.stream_name, self.build_id)

    def url(self):
        self._initialize_nonlocal_data()
        return self._url

    def modification_time(self):
        self._initialize_nonlocal_data()
        return self._modification_time
        
    def _initialize_nonlocal_data(self):
        if not self._url or not self._modification_time:
            stream = OLPCStreams.cached.get_stream_by_name(stream_name)
            build = stream.get_build_by_id(build_id)
            self._url = build.url()
            self._modification_time = build.modification_time()

    def urls_of_ext3_image_and_its_md5(self):
        self._initialize_nonlocal_data()
        if not self._ext3_url_was_looked_for:
            self._ext3_url_was_looked_for = True
            ext3_dir = os.path.join(self._url,"devel_ext3/")
            try:
                page = wget_page(ext3_dir)
                m = re.search('([\w\.-]+\.img\.bz2)\.md5',page)
                ext3_name = m.group(1)
                self._ext3_url = os.path.join(ext3_dir,ext3_name)
            except:
                pass
        url = self._ext3_url
        if url:
            return (url, url+".md5")
        else:
            return (None,None)
